@@ -5,7 +5,7 @@ use warnings;
 package MooX::Press;
 
 our $AUTHORITY = 'cpan:TOBYINK';
-our $VERSION   = '0.005';
+our $VERSION   = '0.006';
 
 use Exporter::Tiny qw(mkopt);
 use Scalar::Util qw(blessed);
@@ -102,7 +102,8 @@ sub import {
 
 sub qualify_name {
 	shift;
-	my ($name, $prefix) = @_;
+	my ($name, $prefix, $parent) = @_;
+	return join("::", $parent, $1) if (defined $parent and $name =~ /^\+(.+)/);
 	return $1 if $name =~ /^::(.+)$/;
 	$prefix ? join("::", $prefix, $name) : $name;
 }
@@ -168,6 +169,7 @@ sub make_type_for_role {
 	my $builder = shift;
 	my ($name, %opts) = @_;
 	return unless $opts{'type_library'};
+	$builder->croak("Roles ($name) cannnot extend things") if $opts{extends};
 	$builder->_make_type($name, %opts, is_role => 1);
 }
 
@@ -181,7 +183,10 @@ sub make_type_for_class {
 sub _make_type {
 	my $builder = shift;
 	my ($name, %opts) = @_;
-	my $qname = $builder->qualify_name($name, $opts{prefix});
+	my @isa = map $builder->qualify_name($_, $opts{prefix}),
+		grep $_,
+		$_merge->(@opts{qw/extends/});
+	my $qname = $builder->qualify_name($name, $opts{prefix}, @isa);
 	
 	my $type_name = $opts{'type_name'} || $builder->type_name($qname, $opts{'prefix'});
 	if ($opts{'type_library'}->can('_mooxpress_add_type')) {
@@ -219,7 +224,11 @@ sub do_coercions_for_class {
 sub _do_coercions {
 	my $builder = shift;
 	my ($name, %opts) = @_;
-	my $qname = $builder->qualify_name($name, $opts{prefix});
+	
+	my @isa = map $builder->qualify_name($_, $opts{prefix}),
+		grep $_,
+		$_merge->(@opts{qw/extends/});
+	my $qname = $builder->qualify_name($name, $opts{prefix}, @isa);
 	
 	if ($opts{coerce}) {
 		my $method_installer = $opts{toolkit_install_methods} || ("install_methods");
@@ -287,7 +296,11 @@ sub make_class {
 sub _make_package {
 	my $builder = shift;
 	my ($name, %opts) = @_;
-	my $qname = $builder->qualify_name($name, $opts{prefix});
+	
+	my @isa = map $builder->qualify_name($_, $opts{prefix}),
+		grep $_,
+		$_merge->(@opts{qw/extends/});
+	my $qname = $builder->qualify_name($name, $opts{prefix}, @isa);
 	
 	if (!exists $opts{factory}) {
 		my $tn = $builder->type_name($qname, $opts{prefix});
@@ -316,9 +329,6 @@ sub _make_package {
 			or $builder->croak("Could not create package $qname: $@");
 	
 		my $method  = $opts{toolkit_extend_class} || ("extend_class_".lc $toolkit);
-		my @isa = map $builder->qualify_name($_, $opts{prefix}),
-			grep $_,
-			$_merge->(@opts{qw/extends/});
 		if (@isa) {
 			$builder->$method($qname, \@isa);
 		}
@@ -434,12 +444,6 @@ sub _make_package {
 				$spec{coerce} = 1;
 			}
 			
-#			if ($attrname eq 'owner') {
-#				use Data::Dumper;
-#				print Dumper(!!$spec{isa}->coercion->can_be_inlined);
-#				die;
-#			}
-			
 			$builder->$method($qname, $attrname, \%spec);
 		}
 	}
@@ -494,19 +498,10 @@ sub _make_package {
 sub make_attribute_moo {
 	my $builder = shift;
 	my ($class, $attribute, $spec) = @_;
-	if ($INC{'Moo/Role.pm'} && 'Moo::Role'->is_role($class)) {
-		# no, I don't like this either
-		($Moo::Role::INFO{$class}{accessor_maker} ||= do {
-			require Method::Generate::Accessor;
-			Method::Generate::Accessor->new
-		})->generate_method($class, $attribute, $spec);
-		push @{$Moo::Role::INFO{$class}{attributes}||=[]}, $attribute, $spec;
-	}
-	else {
-		'Moo'->_constructor_maker_for($class)->register_attribute_specs($attribute, $spec);
-		'Moo'->_accessor_maker_for($class)->generate_method($class, $attribute, $spec);
-	}
-	'Moo'->_maybe_reset_handlemoose($class);
+	my $tracker = ($INC{'Moo/Role.pm'} && 'Moo::Role'->is_role($class))
+		? $Moo::Role::INFO{$class}{exports}
+		: $Moo::MAKERS{$class}{exports};
+	$tracker->{has}->($attribute, %$spec);
 }
 
 sub make_attribute_moose {
@@ -526,8 +521,8 @@ sub make_attribute_mouse {
 sub extend_class_moo {
 	my $builder = shift;
 	my ($class, $isa) = @_;
-	'Moo'->_set_superclasses($class, @$isa);
-	'Moo'->_maybe_reset_handlemoose($class);
+	my $tracker = $Moo::MAKERS{$class}{exports};
+	$tracker->{extends}->(@$isa);
 }
 
 sub extend_class_moose {
@@ -547,9 +542,10 @@ sub extend_class_mouse {
 sub apply_roles_moo {
 	my $builder = shift;
 	my ($class, $roles) = @_;
-	require Moo::Role;
-	'Moo::Role'->apply_roles_to_package($class, @$roles);
-	'Moo'->_maybe_reset_handlemoose($class);
+	my $tracker = ($INC{'Moo/Role.pm'} && 'Moo::Role'->is_role($class))
+		? $Moo::Role::INFO{$class}{exports}
+		: $Moo::MAKERS{$class}{exports};
+	$tracker->{with}->(@$roles);
 }
 
 sub apply_roles_moose {
@@ -824,13 +820,15 @@ The following class options are supported.
 
 =over
 
-=item C<< extends >> I<< (Str) >>
+=item C<< extends >> I<< (Str|ArrayRef[Str]) >>
 
 The parent class for this class.
 
 The prefix is automatically added. Include a leading "::" if you
 don't want the prefix to be added.
 
+Multiple inheritance is supported.
+
 =item C<< with >> I<< (ArrayRef[Str]) >>
 
 Roles for this class to consume.
@@ -1148,15 +1146,21 @@ It allows subclasses to be nested as deep as you like:
   
   package main;
   my $uncle = MyApp->new_human(name => "Bob");
-  $uncle->isa('Human');      # true
-  $uncle->isa('Primate');    # true
-  $uncle->isa('Mammal');     # true
-  $uncle->isa('Animal');     # true
-  $uncle->isa('Bird');       # false
-  $uncle->can('lactate');    # eww, but true
+  $uncle->isa('MyApp::Human');    # true
+  $uncle->isa('MyApp::Primate');  # true
+  $uncle->isa('MyApp::Mammal');   # true
+  $uncle->isa('MyApp::Animal');   # true
+  $uncle->isa('MyApp::Bird');     # false
+  $uncle->can('lactate');         # eww, but true
 
 We just defined a nested heirarchy with ten classes there!
 
+Subclasses can be named with a leading "+" to tell them to use their parent
+class name as a prefix. So, in the example above, if you'd called your
+subclasses "+Mammal", "+Dog", etc, you'd end up with packages like
+"MyApp::Animal::Mammal::Dog". (In cases of multiple inheritance, it uses
+C<< $ISA[0] >>.)
+
 =item C<< factory >> I<< (Str) >>
 
 This is the name for the method installed into the factory package.
@@ -1231,7 +1235,7 @@ exceptions:
 
 =item C<< extends >> I<< (Any) >>
 
-This option is silently ignored.
+This option is disallowed.
 
 =item C<< can >> I<< (HashRef[CodeRef]) >>
 
